/*
 * CovarianceMatrix.h
 *
 *  Created on: Apr 8, 2010
 *      Author: sbak
 */

#ifndef MYCM_H_
#define MYCM_H_

#include "common.h"
#include "Layers.h"


#define METRIC_L1 				0
#define METRIC_L2				(METRIC_L1 + 1)
#define METRIC_RIEMANNIAN 		(METRIC_L2 + 1)
#define METRIC_CHI_SQRT 			(METRIC_RIEMANNIAN + 1)


#define COVARIANCE_PORIKLI					0
#define COVARIANCE_TANGENT_PORIKLI			(COVARIANCE_PORIKLI + 1)
#define COVARIANCE_BROWNIAN					(COVARIANCE_TANGENT_PORIKLI + 1)
#define COVARIANCE_TANGENT_BROWNIAN			(COVARIANCE_BROWNIAN + 1)

/*
 *  returns e^matrix
 */
lti::matrix<float> exp(const lti::matrix<float>& matrix);

/*
 * returns ln(matrix)
 */
lti::matrix<float> ln(const lti::matrix<float>& matrix);

/*
 * matrix is treated as vector and L1 Norm is computed
 * returns sum (matrix);
 */

float L1norm(const lti::matrix<float>& matrix);

/*
 * matrix is treated as vector and L2 (Euclidean Norm) is computed
 * returns sqrt( sum (matrix^2) );
 */
float L2norm(const lti::matrix<float>& matrix);

void norm(lti::matrix<float>& matrix, int metric);


/*
 * creates integral image
 * TODO: in future integral images should be implemented for different types (int/char/etc.) to speed up approach
 */
lti::matrix<float> IntegralImage(const lti::matrix<float>& input);

/*
 * creates integral image of (input1. * input2. )
 */
lti::matrix<float> mIntegralImage(const lti::matrix<float>& input1,const lti::matrix<float>& input2);

class CovarianceMatrix {

	/*
	 * image for which the covariance/covariances will be extracted
	 */
	IplImage* image;
	lti::channel* _mask;
	lti::matrix<float> _maskII;
	bool _INTEGRALS;
	Layers* engine;

	/*
	 * Integral Images of features
	 */
	std::map<int, lti::matrix<float> > II;				//map of integral images for each feature
	std::map< std::string, lti::matrix<float> > mII;	//map of integral images for each combination of features

	void extendImageForGradient();

public:
	CovarianceMatrix(IplImage* img, lti::channel* mask = NULL);
	CovarianceMatrix(const CovarianceMatrix& c);

	virtual ~CovarianceMatrix();

	/*
	 * returns binary mask corresponding to 'non-zero' values in the image
	 */
	static lti::channel getMaskChannel(IplImage* img);

	Layers* getEngine(){return engine;}
	IplImage* getImage();
	IplImage* getRegionImage(CvRect rect);
	int getRegionSize(const CvRect rect);
	/*
	 * creates integral images for features which are selected as TRUE in feature vector
	 * To obtain mapping of features in features vector see featuresNames
	 *
	 * IMPORTANT:
	 * Integral-based implementation computes gradient using whole image in contrary to patch-based implementation producing slightly different output
	 */
	void extractFeatures(std::vector<bool> features, std::vector<CvPoint> points = std::vector<CvPoint>(0));
	void computeIntegrals(std::vector<bool> features);

	/*
	 * computes covariance matrix using features selected as TRUE in features vector
	 */
	lti::matrix<float> getCovarianceMatrix(CvRect rect, std::vector<bool> features);
	/*
		 * computes covariance matrix using features selected as TRUE in features vector
		 */
	lti::matrix<float> getBrownianCovariance(CvRect rect, std::vector<bool> features);
	/*
	 * returns vector containing means of features selected as TRUE in features vector
	 */
	lti::matrix<float> getMeanFeatures(CvRect rect, std::vector<bool> features);

	/*
	 * returns distance or -1 if the eigen values can not be computed
	 */
	static float getEigenDistance(lti::matrix<float>& cov1, lti::matrix<float>& cov2);

	/*
	 * returns distance or -1 if the eigen values can not be computed
	 */
	static float getEigenDistance(CovarianceMatrix* cm1, CovarianceMatrix*  cm2, CvRect rect1, CvRect rect2,std::vector<bool> features);

	/*
	 * compute mean covariance of covariance matrices in Riemanian Manifold space
	 */
	static lti::matrix<float> getMeanCovariance(std::vector< lti::matrix<float> >& matrixes);
	static float getVarianceOfCovariances(std::vector< lti::matrix<float> >& matrixes, lti::matrix<float> mean);

	/*
	 * compute mean covariance of covariance matrices in Tangent space
	*/
	static lti::matrix<float> getMeanVectorIter(std::vector< lti::matrix<float> >& matrixes);
	/*
	 * compute centroid of covariance matrices
	 */
	static lti::matrix<float> getCentroidCovariance(std::vector< lti::matrix<float> >& matrixes);

	/*
	 * compute mean and variance of vectors
	 */
	static lti::matrix<float> getMeanVector(std::vector< lti::matrix<float> >& vectors);
	static float getVarianceVector(std::vector< lti::matrix<float> >& vectors, lti::matrix<float> mean, int metric);
	/*
	 * concerns vectors of means of features (see source code)
	 */
	static float getEuclideanDistance(CovarianceMatrix* cm1, CovarianceMatrix*  cm2, CvRect rect1, CvRect rect2,std::vector<bool> features);

	/*
	 * returns distance based on covariance distance and means of the features
	 * [0] - combined distance
	 * [1] - distance between mean feature vectors
	 * [2] - eigen distance between covariances
	 */
	static float* getCombinedDistance(lti::matrix<float>& cov1, lti::matrix<float>& cov2, lti::matrix<float>& meanF1, lti::matrix<float>& meanF2, float covFactor = 20 , float meanFactor = 100);
	/*
	 * see code
	 */
	static float* getCombinedDistance(CovarianceMatrix* cm1, CovarianceMatrix*  cm2, CvRect rect1,  CvRect rect2, std::vector<bool> features, std::vector<bool> mfeatures, float covFactor = 20 , float meanFactor = 100);

	/*
	 * returns Pearson product-moment correlation coefficients
	 */
	static lti::matrix<float> normalize(lti::matrix<float>& matrix);


	/*
	 *  function which creates covariance matrix from vector of features
	 *  feature vector is represented by lti::matrix
	 *  first mean feature vector is computed and then according to definition of covariance, matrix is calculated (see source code)
	 *  this implementation does not concern FAST computation of covariance (no integrals are computed)
	 */
	static lti::matrix<float> getCovarianceMatrix(std::vector< lti::matrix<float> >& features);

	static std::vector<float> L1v(lti::matrix<float>& vector1, lti::matrix<float>& vector2);
	static float L1(lti::matrix<float>& vector1, lti::matrix<float>& vector2);
	static float L2(lti::matrix<float>& vector1, lti::matrix<float>& vector2);
	static float chi_sqrt(lti::matrix<float>& vector1, lti::matrix<float>& vector2);


	/*
	 * functions which transform matrix to vector and vice versa, here we assume that matrix is symmetric
	 */
	static lti::matrix<float> c2v(lti::matrix<float> c);
	static lti::matrix<float> v2c(lti::matrix<float> v);

	static std::vector<float> _c2v(lti::matrix<float> c);
	static lti::matrix<float> _v2c(std::vector<float> v);


	static float getDistance(lti::matrix<float>& cov1, lti::matrix<float>& cov2, int metric);
	static lti::matrix<float> getMean(std::vector< lti::matrix<float> >& matrixes, int metric);
	static float getVariance(std::vector< lti::matrix<float> >& matrixes, lti::matrix<float> mean, int metric);

	/*
	 * update the matrix - Here, I assume that we apply it only on TANGENT_RIEMANNIAN or EUCLIDEAN
	 * return distance between matrices
	 */
	static float update(lti::matrix<float>& _old, lti::matrix<float> _new, int metric);
	static bool isValid(lti::matrix<float> cov);
};

#endif /* MYCM_H_ */
