#ifndef _IMAGE_H_
#define _IMAGE_H_

#include <stdio.h>
#include <jpixelformat.h>
#include <jpeglib.h>

namespace vsip {

struct rectangle;

// Maximal value for a grey pixel or a single component of a color pixel
#define COMPV_MAX 255
#define BB 100	// ADD BBUI (for binaire bg: white; fg: black with rectangles)

//---------------------------------------------------------------------------
// Struct : _color_pixel
//---------------------------------------------------------------------------

/** _scolor_pixel structure. */
struct _scolor_pixel {		/* Signed component value */
    signed int red:10;
    signed int green:10;
    signed int blue:10;
};


/** A scolor_pixel is the same as _scolor_pixel structure. */
typedef struct _scolor_pixel scolor_pixel;


/** _bg_pixel structure. */
struct _bg_pixel {		/* Double component value */
     unsigned int blue:10;
     unsigned int green:10;
     unsigned int red:10; 
};

/** A bg_pixel is the same as _bg_pixel structure. */
typedef struct _bg_pixel bg_pixel;

/*
 * The following macros can be used with both types of pixels
 */

#define PIXEL_RED(pPix)         ((pPix).red)
#define PIXEL_GREEN(pPix)       ((pPix).green)
#define PIXEL_BLUE(pPix)        ((pPix).blue)

#if (RGB_PIXELSIZE == 4)

#define COLOR_PIXEL_RED(pPix)	((pPix)[3])
#define COLOR_PIXEL_GREEN(pPix)	((pPix)[2])
#define COLOR_PIXEL_BLUE(pPix)	((pPix)[1])
#define COLOR_PIXEL_ALPHA(pPix) ((pPix)[0])   // For COLOR_IMG_4BYTES type

#elif (RGB_PIXELSIZE == 3)

#define COLOR_PIXEL_RED(pPix)	((pPix)[2])
#define COLOR_PIXEL_GREEN(pPix)	((pPix)[1])
#define COLOR_PIXEL_BLUE(pPix)	((pPix)[0])

#else

#error "FATAL WARNING: RGB_PIXELSIZE variable, defined in libjpeg-6b/jmorecfg.h, "
#error "is not 4 nor 3. These two are the only supported values..."

#endif

#define COLOR_PIXEL_HUE(pPix) (pPix[2])
#define COLOR_PIXEL_SAT(pPix) (pPix[1])
#define COLOR_PIXEL_LUM(pPix) (pPix[0])

#define COLOR_PIXEL_VLU(pPix) (pPix[0])

#define NON_BLACK_COLOR(_pix) \
	(COLOR_PIXEL_RED(_pix) + COLOR_PIXEL_GREEN(_pix) + COLOR_PIXEL_BLUE(_pix) > (unsigned int)0)

#define BLACK_COLOR(_pix) \
	(COLOR_PIXEL_RED(_pix) == 0 && COLOR_PIXEL_GREEN(_pix) == 0 && COLOR_PIXEL_BLUE(_pix) == 0)

#define NON_BLACK_PCOLOR(_ppix)	(*((unsigned int *)_ppix) > (unsigned int)0)
/** @} */ 

/**
 * constants used to designate a single band in an image
 */
typedef enum {
    RED_BAND = 0x1,
    GREEN_BAND = 0x2,
    BLUE_BAND = 0x4
} band_t;

#define HUE_BAND	RED_BAND
#define SAT_BAND	GREEN_BAND
#define LUM_BAND	BLUE_BAND
#define VLU_BAND	BLUE_BAND


typedef unsigned char grey_pixel;

#define PIXEL_VAL(_pix)		(_pix)


typedef float float_pixel;

typedef int int_pixel;

/** image type. */
typedef enum {
    UNKN_IMG = 0,
    COLOR_IMG_3BYTES = -1,	/**< Color image with three component (unsigned)per pixel */
    SCOLOR_IMG = -2,	/**< Color image with 3 bytes per pixel */
    GREY_IMG = -3,	/**< Grey level image: pixel are currently unsigned int values */
    FLOAT_IMG = -4,	/**< floating point images */
    INT_IMG = -5,	/**< signed integer images */
    BG_IMG = -6,        /**< double images */
    COLOR_IMG_4BYTES = -7     /**< Color image with 4 bytes per pixel */
} image_type;

#define RGB_IMG	COLOR_IMG_3BYTES
#define HSL_IMG	COLOR_IMG_3BYTES


#define IMAGE_WIDTH_MAX		1024

/** image structure. */
struct image 
{
  image_type type;
  unsigned int width;
  unsigned int height;
  unsigned int data_size;
  size_t pixel_size;	 /** filled with sizeof(pixel type)/sizeof(char) */
  int x_offset;
  int y_offset;
  union 
  {
    unsigned char *c_data;  // This field is used by both images COLOR_IMG_3BYTES and COLOR_IMG_4BYTES
    scolor_pixel *sc_data;
    grey_pixel *g_data;
    float_pixel *f_data;
    int_pixel *i_data;
    char *u_data;
    bg_pixel *bg_data;
  } bitmap;
  char * jpeg_marker_string; /** filled with marker APP0 if the source JPEG image had one */
  struct image *next;
};

typedef struct image *image_t;

/** @{ */
/**
 * Macro for getting image attribut.
 */
#define IMAGE_TYPE(_img)		((_img)->type)
#define IMAGE_WIDTH(_img)		((_img)->width)
#define IMAGE_HEIGHT(_img)		((_img)->height)
#define IMAGE_COLOR_BITMAP(_img)	((_img)->bitmap.c_data)
#define IMAGE_SCOLOR_BITMAP(_img)	((_img)->bitmap.sc_data)
#define IMAGE_GREY_BITMAP(_img)		((_img)->bitmap.g_data)
#define IMAGE_FLOAT_BITMAP(_img)	((_img)->bitmap.f_data)
#define IMAGE_INT_BITMAP(_img)		((_img)->bitmap.i_data)
#define IMAGE_BG_BITMAP(_img)		((_img)->bitmap.bg_data)
#define IMAGE_JPEG_APP0(_img)           ((_img)->jpeg_marker_string)

#if (RGB_PIXELSIZE == 3)

#define IMAGE_COLOR_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.c_data + 3*((_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width)))

#else

#define IMAGE_COLOR_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.c_data + 4*((_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width)))

#endif

#define IMAGE_SCOLOR_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.sc_data + (_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width))
#define IMAGE_GREY_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.g_data + (_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width))
#define IMAGE_FLOAT_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.f_data + (_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width))
#define IMAGE_INT_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.i_data + (_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width))
#define IMAGE_CHAR_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.u_data \
	 + ((_x) - (_img)->x_offset \
	    + (((_y) - (_img)->y_offset) * (_img)->width)) * (_img)->pixel_size)
#define IMAGE_BG_PIXPTR(_img, _x, _y) \
	((_img)->bitmap.bg_data + (_x) - (_img)->x_offset \
	 + (((_y) - (_img)->y_offset) * (_img)->width))

#define IMAGE_NEXT(_img)		((_img)->next)
/** @} */

/**
 * A macro to set the offset of an image
 */
#define IMAGE_SET_OFFSET(img, x_off, y_off)	\
	if ((img) != NULL) { \
    	  (img)->x_offset = x_off; \
	  (img)->y_offset = y_off; \
	}

/*
 * For the fourth argument to the function copy_image_content()
 */

typedef enum {
    COPY_AT_POSITION,	/* copy content at the same position */
    COPY_AT_ORIGIN	/* copy content at the origin of the output image */
} copy_type_t;


/*
 * For handling JPEG compression/decompression in a thread-like mode
 */

typedef struct {
  struct jpeg_decompress_struct *c_decomp_info;
  struct jpeg_error_mgr *jerr;
  struct jpeg_compress_struct *c_comp_info;
  char * buffer;
  int buffer_size;
  int data_size;
} jpeg_manager_st;

typedef jpeg_manager_st * jpeg_manager_t;

/*type de ligne*/
typedef enum {
    HOR,
    VERT
} line_type;

/**
 * builds and initializes an image of the required type (COLOR, GREY or FLOAT),
 * with the required height (h) and width (w). The initialization is performed
 * by allocating a big enough space for the content of the image.
 * Returns the new image itself, or NULL if an error occurred.
 */
image_t make_image(image_type type, unsigned int w, unsigned int h);

/**
 * performs the initialization of an image. The space requirement are checked
 * and enough space is allocated if necessary. All previously allocated space
 * for this image is garbagged if it is needed (ie. if the space is not
 * sufficient with respect to the new size and type of the image.
 * Returns a negative number on error, 0 otherwise.
 */
int initialize_image(image_t img, image_type type,
		     unsigned int w, unsigned int h);

int _write_header(image_t img, FILE *out);
int _write_color_data(image_t img, FILE *out);
int _write_grey_data(image_t img, FILE *out);
 int _write_float_data(image_t img,FILE *out);

/** Save a image to .PPM file */
int save_image(image_t img, const char *file_name);

/** Save a gradient to .PPM files */
int save_gradient(image_t gradx,image_t grady, const char *file_name);

/**
 * Clear the whole content of an image
 */
int clear_image(image_t img);

/**
 * Clear the subpart of the image as indicated by the rectangle given as the
 * second argument
 */
int clear_image_area(image_t img, struct rectangle* rect);

/**
 * Destroy an image and frees all space previously allocated for its content
 */
int destroy_image(image_t img);

/**
 * Returns a copy of the whole image. 
 */
image_t copy_whole_image(image_t img);

/**
 * Transfers the content of the input image in the indicated rectangle to the
 * output image. depending on the fourth argument, different policies of copying
 * are taken.
 */
int copy_image_content(image_t from_img, struct rectangle* rect,
		       image_t to_img, copy_type_t copy_type);

/**
 * Initializes the given rectangle so that it corresponds to the bounding
 * rectangle of the whole image
 */
int init_bounding_box(struct rectangle* rect, image_t img);

/**
 * Performs a general convolution of the input image using the given mask. The
 * mask is an array of integer values of size mask_size * mask_size, where
 * the values are put row after row. The output is placed in the same image.
 */
int apply_mask_in_place(image_t img, struct rectangle* rect,
			int mask_size, int *mask_data);
/**
 * Performs a general convolution of the input image using the given mask. The
 * mask is an array of integer values of size mask_size * mask_size, where
 * the values are put row after row. The output is placed in the output image.
 */
int apply_mask(image_t in_img, struct rectangle* rect,
	       int mask_size, int *mask_data, image_t out_img);

/** Not more used */
int update(image_t img, image_t new_img, float gain, struct rectangle* rect);

/** Not more used */
int color_thresholded_update(image_t img, image_t new_img, float gain,
			     int red_th, int green_th, int blue_th,
			     struct rectangle* rect);

/** Not more used */
int grey_thresholded_update(image_t img, image_t new_img, float gain,
			    int threshold, struct rectangle* rect);


/** Not more used */
int abs_difference(image_t img1, image_t img2, struct rectangle* rect,
		   image_t out_img);

/** Not more used */
int sign_difference(image_t img1, image_t img2, struct rectangle* rect,
		   image_t out_img);
#define TH_OR	0x01
#define TH_AND	0x02

/** Not more used */
int color_thresholding(image_t in_img,
		       unsigned int red_th, unsigned int green_th,
		       unsigned int blue_th,
		       struct rectangle* rect,
		       band_t bands, int combine,
		       image_t out_img);

/** Not more used */
int grey_thresholding(image_t in_img, unsigned int th, struct rectangle* rect,
		      image_t out_img);

/** Not more used */
int float_thresholding(image_t in_img, float thr,
		       struct rectangle* rect, image_t out_img);

/** Not more used */
int int_thresholding(image_t in_img, unsigned int th, struct rectangle* rect,
		     image_t out_img);

/** Not more used */
image_t convert_to_grey(image_t in_img, struct rectangle* rect);


/** Draw a rectangle on image. */
int draw_rectangle(struct rectangle* rect, image_t image);

/** Draw a rectangle with color on image. */
int draw_color_rectangle(struct rectangle* rect, image_t image, int r, int g, int b);

/*!
 *\fn int bb_draw_color_line(int x1,int y1,line_type ltype,image_t image,int r,int g,int b);
 *\brief this function draw a line in image
 *
 *
 *\param x1 coordinate of start point of the line 
 *\param x2 coordinate of end point of the line 
 *\param y where is the line
 *\param ltype HOR or VERT
 *\param image The image where the line is draw
 *\param r componet of red [0..255]
 *\param g component of green [0..255]
 *\param b conmponent of blue [0..255]
 *\return 0 if success, -1 otherwise
 */
  int bb_draw_color_line(int x1,int x2,int y, line_type ltype,image_t image,int r,int g,int b);

/*!
 *\fn int bb_draw_color_line_diag(int x1,int y1,int x2,int y2,image_t image,int r,int g,int b);
 *\brief this function draw a line in image
 *
 *
 *\param x1,y1 coordinates of start point of the line 
 *\param x2,y2 coordinates of end point of the line 
 *\param image The image where the line is draw
 *\param r componet of red [0..255]
 *\param g component of green [0..255]
 *\param b conmponent of blue [0..255]
 *\return 0 if success, -1 otherwise
 */
  int bb_draw_color_line_diag(int x1,int y1,int x2,int y2,image_t image,int r,int g,int b);


/** Not more used */
int copy_to_bg (image_t from_img, image_t to_bg, struct rectangle* bbox);

  /** Not more used */
int copy_from_bg (image_t from_bg, image_t to_img, struct rectangle* bbox);

/** Not more used */
image_t convert_to_1byte(image_t in_img, struct rectangle* rect, image_t out_img);
/* This function takes as input a color binary image (in_img) and convert it
   to a binary grey image (out_img, or given back by the function).
   If rect == NULL, the whole image is converted, otherwise only the image part
   surrended by rect is converted.
   If out_img is != NULL (image still allocated), the function works using this
   image (warning! Any control isn't made over the dimensions of the out_img and
   its memory space!). If out_img == NULL, out_img is created.
*/

/** Not more used */
image_t convert_to_4bytes(image_t in_img, struct rectangle* rect, image_t out_img);
/* This function takes as input a grey binary image (in_img) and convert it
   to a binary color image (this function is only useful when it is necessary
   to display the image, because _nr_display doesn't handle grey images).
   If rect == NULL, the whole image is converted, otherwise only the image part
   surrended by rect is converted.
   If out_img is != NULL (image still allocated), the function works using this
   image (warning! Any control isn't made over the dimensions of the out_img and
   its memory space!). If out_img == NULL, out_img is created.
*/

image_t image_data_encapsulate(char * raw_image_ptr, int h, int w, image_t new_image, image_type type);
/* This function takes as input an array of char (raw_image_ptr) containing
   raw image data, in this format: 
   for type = COLOR_IMG_4BYTES, raw_image_ptr[0-3] = dummy (D), blue (B),
       green (G) and red (R) components of the first pixel, raw_image_ptr[4-7]
       DBGR components of the second pixel, etc.
   for type = GREY_IMG, raw_image_ptr[0] = first pixel, raw_image_ptr[1] = 
       second pixel and so on.
   for type = COLOR_IMG_3BYTES, raw_image_ptr[0-2] = red (B), green (G) and blue (R) 
       components of the first pixel, raw_image_ptr[3-5] BGR components of the second pixel, etc.
   
   h and w are height and width of the image. Data array raw_image_ptr has
   to be greater then h*w*sizeof(XXX_pixel) (XXX_pixel = color_pixel or grey_pixel)
   
   new_image is an image_t image (or NULL) where to write raw_image_ptr as
   image data. If NULL, a new image is allocated. If != NULL, memory space is
   checked and, if necessary, old image data is erased.

   WARNING!!!! If you don't want to free data space of an existing new_image,
   you have to set IMAGE_COLOR_BITMAP(new_image)=NULL before calling this function!
   (And in this case pay attention to memory leakages possibility).
   
   type is COLOR_IMG_3BYTES, COLOR_IMG_4BYTES or GRAY_IMG (other types are not yet supported)

   Function gives back an image_t standard structure, completed with missing
   data.

   If h=0, w=0 or raw_image_ptr = NULL function gives back NULL 

   ATTENTION: type MUST be coherent with the lenght and the nature of the raw_image_ptr. If this
   last one is DRGBDRGB, type will be COLOR_IMG_4BYTES, if it is RGBRGB it will be COLOR_IMG_3BYTES*/

void check_RGB_channels_extension(image_t img, int * max_channel, int * min_channel, int n_channel);
/* This function takes as input a color image and a color channel number (R = 0; G = 1; B = others).
   It gives back the maximum and minimum values of this channel over the whole image.
   If an error occurs, it gives back -1 and -1. */

void merge_binary_images(image_t io_img1, image_t i_img2);
/* This function merge the two images io_img1 and i_img2 into io_img1.
   io_img1 must be a grey binary image GREY_IMG (0 or 255 only allowed) while i_img2 can be a color 
   image COLOR_IMG_3BYTES (if there is a colour, it is considered as black)
   or a grey image GREY_IMG (if there a grey level different than 255, it is considered as black).*/

image_t merge_binary_images_color(image_t io_img1, image_t io_img2);
   
} // end namespace vsip
	 
#endif /* _IMAGE_H_ */
